@@ -3,13 +3,14 @@
 
 #include <err.h>
 #include <errno.h>
-#include <inttypes.h>
 #include <libgen.h>
 #include <limits.h>
 #include <locale.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <time.h>
 #include <unistd.h>
 #include <wchar.h>
 
@@ -17,6 +18,10 @@
 
 #include "compat.h"
 
+#define LEN(s)    (sizeof(s)/sizeof(*s))
+#define PAD_TRUNCATE_SYMBOL    "\xe2\x80\xa6" /* symbol: "ellipsis" */
+#define UTF_INVALID_SYMBOL     "\xef\xbf\xbd" /* symbol: "replacement" */
+
 struct deltainfo {
 	git_patch *patch;
 
@@ -49,8 +54,15 @@ struct commitinfo {
 	size_t ndeltas;
 };
 
+/* reference and associated data for sorting */
+struct referenceinfo {
+	struct git_reference *ref;
+	struct commitinfo *ci;
+};
+
 static git_repository *repo;
 
+static const char *baseurl = ""; /* base URL to make absolute RSS/Atom URI */
 static const char *relpath = "";
 static const char *repodir;
 
@@ -58,48 +70,74 @@ static char *name = "";
 static char *strippedname = "";
 static char description[255];
 static char cloneurl[1024];
-static int haslicense, hasreadme, hassubmodules;
+static char *submodules;
+static char *licensefiles[] = { "HEAD:LICENSE", "HEAD:LICENSE.md", "HEAD:COPYING" };
+static char *license;
+static char *readmefiles[] = { "HEAD:README", "HEAD:README.md" };
+static char *readme;
+static long long nlogcommits = -1; /* < 0 indicates not used */
 
 /* cache */
 static git_oid lastoid;
-static char lastoidstr[GIT_OID_HEXSZ + 2]; /* id + newline + nul byte */
+static char lastoidstr[GIT_OID_HEXSZ + 2]; /* id + newline + NUL byte */
 static FILE *rcachefp, *wcachefp;
 static const char *cachefile;
 
-#ifndef USE_PLEDGE
-#define pledge(p1,p2) 0
-#endif
-
-/* format `len' columns of characters. If string is shorter pad the rest
+/* Format `len' columns of characters. If string is shorter pad the rest
  * with characters `pad`. */
 int
 utf8pad(char *buf, size_t bufsiz, const char *s, size_t len, int pad)
 {
-	wchar_t w;
+	wchar_t wc;
 	size_t col = 0, i, slen, siz = 0;
-	int rl, wc;
+	int inc, rl, w;
 
-	if (!len)
+	if (!bufsiz)
 		return -1;
+	if (!len) {
+		buf[0] = '\0';
+		return 0;
+	}
 
 	slen = strlen(s);
-	for (i = 0; i < slen && col < len + 1; i += rl) {
-		if ((rl = mbtowc(&w, &s[i], slen - i < 4 ? slen - i : 4)) <= 0)
-			break;
-		if ((wc = wcwidth(w)) == -1)
-			wc = 1;
-		col += wc;
-		if (col >= len && s[i + rl]) {
+	for (i = 0; i < slen; i += inc) {
+		inc = 1; /* next byte */
+		if ((unsigned char)s[i] < 32)
+			continue;
+
+		rl = mbtowc(&wc, &s[i], slen - i < 4 ? slen - i : 4);
+		inc = rl;
+		if (rl < 0) {
+			mbtowc(NULL, NULL, 0); /* reset state */
+			inc = 1; /* invalid, seek next byte */
+			w = 1; /* replacement char is one width */
+		} else if ((w = wcwidth(wc)) == -1) {
+			continue;
+		}
+
+		if (col + w > len || (col + w == len && s[i + inc])) {
 			if (siz + 4 >= bufsiz)
 				return -1;
-			memcpy(&buf[siz], "\xe2\x80\xa6", 4);
-			return 0;
+			memcpy(&buf[siz], PAD_TRUNCATE_SYMBOL, sizeof(PAD_TRUNCATE_SYMBOL) - 1);
+			siz += sizeof(PAD_TRUNCATE_SYMBOL) - 1;
+			buf[siz] = '\0';
+			col++;
+			break;
+		} else if (rl < 0) {
+			if (siz + 4 >= bufsiz)
+				return -1;
+			memcpy(&buf[siz], UTF_INVALID_SYMBOL, sizeof(UTF_INVALID_SYMBOL) - 1);
+			siz += sizeof(UTF_INVALID_SYMBOL) - 1;
+			buf[siz] = '\0';
+			col++;
+			continue;
 		}
-		if (siz + rl + 1 >= bufsiz)
+		if (siz + inc + 1 >= bufsiz)
 			return -1;
-		memcpy(&buf[siz], &s[i], rl);
-		siz += rl;
+		memcpy(&buf[siz], &s[i], inc);
+		siz += inc;
 		buf[siz] = '\0';
+		col += w;
 	}
 
 	len -= col;
@@ -119,7 +157,7 @@ joinpath(char *buf, size_t bufsiz, const char *path, const char *path2)
 
 	r = snprintf(buf, bufsiz, "%s%s%s",
 		path, path[0] && path[strlen(path) - 1] != '/' ? "/" : "", path2);
-	if (r == -1 || (size_t)r >= bufsiz)
+	if (r < 0 || (size_t)r >= bufsiz)
 		errx(1, "path truncated: '%s%s%s'",
 			path, path[0] && path[strlen(path) - 1] != '/' ? "/" : "", path2);
 }
@@ -130,7 +168,7 @@ deltainfo_free(struct deltainfo *di)
 	if (!di)
 		return;
 	git_patch_free(di->patch);
-	di->patch = NULL;
+	memset(di, 0, sizeof(*di));
 	free(di);
 }
 
@@ -138,6 +176,8 @@ int
 commitinfo_getstats(struct commitinfo *ci)
 {
 	struct deltainfo *di;
+	git_diff_options opts;
+	git_diff_find_options fopts;
 	const git_diff_delta *delta;
 	const git_diff_hunk *hunk;
 	const git_diff_line *line;
@@ -145,6 +185,30 @@ commitinfo_getstats(struct commitinfo *ci)
 	size_t ndeltas, nhunks, nhunklines;
 	size_t i, j, k;
 
+	if (git_tree_lookup(&(ci->commit_tree), repo, git_commit_tree_id(ci->commit)))
+		goto err;
+	if (!git_commit_parent(&(ci->parent), ci->commit, 0)) {
+		if (git_tree_lookup(&(ci->parent_tree), repo, git_commit_tree_id(ci->parent))) {
+			ci->parent = NULL;
+			ci->parent_tree = NULL;
+		}
+	}
+
+	git_diff_init_options(&opts, GIT_DIFF_OPTIONS_VERSION);
+	opts.flags |= GIT_DIFF_DISABLE_PATHSPEC_MATCH |
+	              GIT_DIFF_IGNORE_SUBMODULES |
+	              GIT_DIFF_INCLUDE_TYPECHANGE;
+	if (git_diff_tree_to_tree(&(ci->diff), repo, ci->parent_tree, ci->commit_tree, &opts))
+		goto err;
+
+	if (git_diff_find_init_options(&fopts, GIT_DIFF_FIND_OPTIONS_VERSION))
+		goto err;
+	/* find renames and copies, exact matches (no heuristic) for renames. */
+	fopts.flags |= GIT_DIFF_FIND_RENAMES | GIT_DIFF_FIND_COPIES |
+	               GIT_DIFF_FIND_EXACT_MATCH_ONLY;
+	if (git_diff_find_similar(ci->diff, &fopts))
+		goto err;
+
 	ndeltas = git_diff_num_deltas(ci->diff);
 	if (ndeltas && !(ci->deltas = calloc(ndeltas, sizeof(struct deltainfo *))))
 		err(1, "calloc");
@@ -152,6 +216,7 @@ commitinfo_getstats(struct commitinfo *ci)
 	for (i = 0; i < ndeltas; i++) {
 		if (git_patch_from_diff(&patch, ci->diff, i))
 			goto err;
+
 		if (!(di = calloc(1, sizeof(struct deltainfo))))
 			err(1, "calloc");
 		di->patch = patch;
@@ -186,6 +251,14 @@ commitinfo_getstats(struct commitinfo *ci)
 	return 0;
 
 err:
+	git_diff_free(ci->diff);
+	ci->diff = NULL;
+	git_tree_free(ci->commit_tree);
+	ci->commit_tree = NULL;
+	git_tree_free(ci->parent_tree);
+	ci->parent_tree = NULL;
+	git_commit_free(ci->parent);
+	ci->parent = NULL;
 	if (ci->deltas)
 		for (i = 0; i < ci->ndeltas; i++)
 			deltainfo_free(ci->deltas[i]);
@@ -210,12 +283,12 @@ commitinfo_free(struct commitinfo *ci)
 		for (i = 0; i < ci->ndeltas; i++)
 			deltainfo_free(ci->deltas[i]);
 	free(ci->deltas);
-	ci->deltas = NULL;
 	git_diff_free(ci->diff);
 	git_tree_free(ci->commit_tree);
 	git_tree_free(ci->parent_tree);
 	git_commit_free(ci->commit);
 	git_commit_free(ci->parent);
+	memset(ci, 0, sizeof(*ci));
 	free(ci);
 }
 
@@ -223,7 +296,6 @@ struct commitinfo *
 commitinfo_getbyoid(const git_oid *id)
 {
 	struct commitinfo *ci;
-	git_diff_options opts;
 
 	if (!(ci = calloc(1, sizeof(struct commitinfo))))
 		err(1, "calloc");
@@ -240,20 +312,6 @@ commitinfo_getbyoid(const git_oid *id)
 	ci->summary = git_commit_summary(ci->commit);
 	ci->msg = git_commit_message(ci->commit);
 
-	if (git_tree_lookup(&(ci->commit_tree), repo, git_commit_tree_id(ci->commit)))
-		goto err;
-	if (!git_commit_parent(&(ci->parent), ci->commit, 0)) {
-		if (git_tree_lookup(&(ci->parent_tree), repo, git_commit_tree_id(ci->parent))) {
-			ci->parent = NULL;
-			ci->parent_tree = NULL;
-		}
-	}
-
-	git_diff_init_options(&opts, GIT_DIFF_OPTIONS_VERSION);
-	opts.flags |= GIT_DIFF_DISABLE_PATHSPEC_MATCH;
-	if (git_diff_tree_to_tree(&(ci->diff), repo, ci->parent_tree, ci->commit_tree, &opts))
-		goto err;
-
 	return ci;
 
 err:
@@ -262,13 +320,110 @@ err:
 	return NULL;
 }
 
+int
+refs_cmp(const void *v1, const void *v2)
+{
+	const struct referenceinfo *r1 = v1, *r2 = v2;
+	time_t t1, t2;
+	int r;
+
+	if ((r = git_reference_is_tag(r1->ref) - git_reference_is_tag(r2->ref)))
+		return r;
+
+	t1 = r1->ci->author ? r1->ci->author->when.time : 0;
+	t2 = r2->ci->author ? r2->ci->author->when.time : 0;
+	if ((r = t1 > t2 ? -1 : (t1 == t2 ? 0 : 1)))
+		return r;
+
+	return strcmp(git_reference_shorthand(r1->ref),
+	              git_reference_shorthand(r2->ref));
+}
+
+int
+getrefs(struct referenceinfo **pris, size_t *prefcount)
+{
+	struct referenceinfo *ris = NULL;
+	struct commitinfo *ci = NULL;
+	git_reference_iterator *it = NULL;
+	const git_oid *id = NULL;
+	git_object *obj = NULL;
+	git_reference *dref = NULL, *r, *ref = NULL;
+	size_t i, refcount;
+
+	*pris = NULL;
+	*prefcount = 0;
+
+	if (git_reference_iterator_new(&it, repo))
+		return -1;
+
+	for (refcount = 0; !git_reference_next(&ref, it); ) {
+		if (!git_reference_is_branch(ref) && !git_reference_is_tag(ref)) {
+			git_reference_free(ref);
+			ref = NULL;
+			continue;
+		}
+
+		switch (git_reference_type(ref)) {
+		case GIT_REF_SYMBOLIC:
+			if (git_reference_resolve(&dref, ref))
+				goto err;
+			r = dref;
+			break;
+		case GIT_REF_OID:
+			r = ref;
+			break;
+		default:
+			continue;
+		}
+		if (!git_reference_target(r) ||
+		    git_reference_peel(&obj, r, GIT_OBJ_ANY))
+			goto err;
+		if (!(id = git_object_id(obj)))
+			goto err;
+		if (!(ci = commitinfo_getbyoid(id)))
+			break;
+
+		if (!(ris = reallocarray(ris, refcount + 1, sizeof(*ris))))
+			err(1, "realloc");
+		ris[refcount].ci = ci;
+		ris[refcount].ref = r;
+		refcount++;
+
+		git_object_free(obj);
+		obj = NULL;
+		git_reference_free(dref);
+		dref = NULL;
+	}
+	git_reference_iterator_free(it);
+
+	/* sort by type, date then shorthand name */
+	qsort(ris, refcount, sizeof(*ris), refs_cmp);
+
+	*pris = ris;
+	*prefcount = refcount;
+
+	return 0;
+
+err:
+	git_object_free(obj);
+	git_reference_free(dref);
+	commitinfo_free(ci);
+	for (i = 0; i < refcount; i++) {
+		commitinfo_free(ris[i].ci);
+		git_reference_free(ris[i].ref);
+	}
+	free(ris);
+
+	return -1;
+}
+
 FILE *
-efopen(const char *name, const char *flags)
+efopen(const char *filename, const char *flags)
 {
 	FILE *fp;
 
-	if (!(fp = fopen(name, flags)))
-		err(1, "fopen: '%s'", name);
+	if (!(fp = fopen(filename, flags)))
+		err(1, "fopen: '%s'", filename);
 
 	return fp;
 }
@@ -286,7 +441,7 @@ xmlencode(FILE *fp, const char *s, size_t len)
 		case '\'': fputs("&#39;",  fp); break;
 		case '&':  fputs("&amp;",  fp); break;
 		case '"':  fputs("&quot;", fp); break;
-		default:   fputc(*s, fp);
+		default:   putc(*s, fp);
 		}
 	}
 }
@@ -298,21 +453,16 @@ gphtextnl(FILE *fp, const char *s, size_t len)
 	size_t i, n = 0;
 
 	for (i = 0; s[i] && i < len; i++) {
-		if (s[i] == '\n')
-			n = 0;
-
 		/* escape with 't' at the start of a line */
-		if (!n && (s[i] == 't' || s[i] == '[')) {
-			fputc('t', fp);
-			n = 1;
-		}
+		if (!n && (s[i] == 't' || s[i] == '['))
+			putc('t', fp);
 
 		switch (s[i]) {
+		case '\t': fputs("        ", fp);
 		case '\r': break;
-		case '\t': fputs("        ", fp); break;
-		default: fputc(s[i], fp);
+		default: putc(s[i], fp);
 		}
-		n++;
+		n = (s[i] != '\n');
 	}
 }
 
@@ -332,7 +482,7 @@ gphtext(FILE *fp, const char *s, size_t len)
 			fputs("        ", fp);
 			break;
 		default:
-			fputc(*s, fp);
+			putc(*s, fp);
 			break;
 		}
 	}
@@ -356,7 +506,7 @@ gphlink(FILE *fp, const char *s, size_t len)
 			fputs("\\|", fp);
 			break;
 		default:
-			fputc(*s, fp);
+			putc(*s, fp);
 			break;
 		}
 	}
@@ -432,7 +582,7 @@ printtimeshort(FILE *fp, const git_time *intime)
 void
 writeheader(FILE *fp, const char *title)
 {
-	fputc('t', fp);
+	putc('t', fp);
 	gphtext(fp, title, strlen(title));
 	if (title[0] && strippedname[0])
 		fputs(" - ", fp);
@@ -451,12 +601,15 @@ writeheader(FILE *fp, const char *title)
 	fprintf(fp, "[1|Log|%s/log.gph|server|port]\n", relpath);
 	fprintf(fp, "[1|Files|%s/files.gph|server|port]\n", relpath);
 	fprintf(fp, "[1|Refs|%s/refs.gph|server|port]\n", relpath);
-	if (hassubmodules)
-		fprintf(fp, "[1|Submodules|%s/file/.gitmodules.gph|server|port]\n", relpath);
-	if (hasreadme)
-		fprintf(fp, "[1|README|%s/file/README.gph|server|port]\n", relpath);
-	if (haslicense)
-		fprintf(fp, "[1|LICENSE|%s/file/LICENSE.gph|server|port]\n", relpath);
+	if (submodules)
+		fprintf(fp, "[1|Submodules|%s/file/%s.gph|server|port]\n",
+		        relpath, submodules);
+	if (readme)
+		fprintf(fp, "[1|README|%s/file/%s.gph|server|port]\n",
+		        relpath, readme);
+	if (license)
+		fprintf(fp, "[1|LICENSE|%s/file/%s.gph|server|port]\n",
+		        relpath, license);
 	fputs("---\n", fp);
 }
 
@@ -465,16 +618,16 @@ writefooter(FILE *fp)
 {
 }
 
-int
+size_t
 writeblobgph(FILE *fp, const git_blob *blob)
 {
-	size_t n = 0, i, j, prev;
-	const char *nfmt = "%6d ";
+	size_t n = 0, i, j, len, prev;
+	const char *nfmt = "%6zu ";
 	const char *s = git_blob_rawcontent(blob);
-	git_off_t len = git_blob_rawsize(blob);
 
+	len = git_blob_rawsize(blob);
 	if (len > 0) {
-		for (i = 0, prev = 0; i < (size_t)len; i++) {
+		for (i = 0, prev = 0; i < len; i++) {
 			if (s[i] != '\n')
 				continue;
 			n++;
@@ -483,7 +636,7 @@ writeblobgph(FILE *fp, const git_blob *blob)
 				switch (s[j]) {
 				case '\r': break;
 				case '\t': fputs("        ", fp); break;
-				default: fputc(s[j], fp);
+				default: putc(s[j], fp);
 				}
 			}
 			prev = i + 1;
@@ -496,7 +649,7 @@ writeblobgph(FILE *fp, const git_blob *blob)
 				switch (s[j]) {
 				case '\r': break;
 				case '\t': fputs("        ", fp); break;
-				default: fputc(s[j], fp);
+				default: putc(s[j], fp);
 				}
 			}
 		}
@@ -525,12 +678,12 @@ printcommit(FILE *fp, struct commitinfo *ci)
 		fputs("|server|port]\n", fp);
 		fputs("Date:   ", fp);
 		printtime(fp, &(ci->author->when));
-		fputc('\n', fp);
+		putc('\n', fp);
 	}
 	if (ci->msg) {
-		fputc('\n', fp);
+		putc('\n', fp);
 		gphtextnl(fp, ci->msg, strlen(ci->msg));
-		fputc('\n', fp);
+		putc('\n', fp);
 	}
 }
 
@@ -543,6 +696,7 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 	git_patch *patch;
 	size_t nhunks, nhunklines, changed, add, del, total, i, j, k;
 	char buf[256], filename[256], linestr[32];
+	int c;
 
 	printcommit(fp, ci);
 
@@ -561,6 +715,17 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 	fputs("Diffstat:\n", fp);
 	for (i = 0; i < ci->ndeltas; i++) {
 		delta = git_patch_get_delta(ci->deltas[i]->patch);
+
+		switch (delta->status) {
+		case GIT_DELTA_ADDED:      c = 'A'; break;
+		case GIT_DELTA_COPIED:     c = 'C'; break;
+		case GIT_DELTA_DELETED:    c = 'D'; break;
+		case GIT_DELTA_MODIFIED:   c = 'M'; break;
+		case GIT_DELTA_RENAMED:    c = 'R'; break;
+		case GIT_DELTA_TYPECHANGE: c = 'T'; break;
+		default:                   c = ' '; break;
+		}
+
 		if (strcmp(delta->old_file.path, delta->new_file.path)) {
 			snprintf(filename, sizeof(filename), "%s -> %s",
 				delta->old_file.path, delta->new_file.path);
@@ -568,7 +733,7 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 		} else {
 			utf8pad(buf, sizeof(buf), delta->old_file.path, 35, ' ');
 		}
-		fputs("  ", fp);
+		fprintf(fp, "  %c ", c);
 		gphtext(fp, buf, strlen(buf));
 
 		add = ci->deltas[i]->addcount;
@@ -601,9 +766,13 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 		patch = ci->deltas[i]->patch;
 		delta = git_patch_get_delta(patch);
 		/* NOTE: only links to new path */
-		fprintf(fp, "[1|diff --git a/%s b/%s",
-			delta->old_file.path, delta->new_file.path);
-		fprintf(fp, "|%s/file/%s.gph|server|port]\n", relpath, delta->new_file.path);
+		fputs("[1|diff --git a/", fp);
+		gphlink(fp, delta->old_file.path, strlen(delta->old_file.path));
+		fputs(" b/", fp);
+		gphlink(fp, delta->new_file.path, strlen(delta->new_file.path));
+		fprintf(fp, "|%s/file/", relpath);
+		gphlink(fp, delta->new_file.path, strlen(delta->new_file.path));
+		fputs(".gph|server|port]\n", fp);
 
 		/* check binary data */
 		if (delta->flags & GIT_DIFF_FLAG_BINARY) {
@@ -616,9 +785,9 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 			if (git_patch_get_hunk(&hunk, &nhunklines, patch, j))
 				break;
 
-			fputc('t', fp);
+			putc('t', fp);
 			gphtext(fp, hunk->header, hunk->header_len);
-			fputc('\n', fp);
+			putc('\n', fp);
 
 			for (k = 0; ; k++) {
 				if (git_patch_get_line_in_hunk(&line, patch, j, k))
@@ -630,7 +799,7 @@ printshowfile(FILE *fp, struct commitinfo *ci)
 				else
 					fputs(" ", fp);
 				gphtext(fp, line->content, line->content_len);
-				fputc('\n', fp);
+				putc('\n', fp);
 			}
 		}
 	}
@@ -647,13 +816,12 @@ writelogline(FILE *fp, struct commitinfo *ci)
 	else
 		fputs("                ", fp);
 	fputs("  ", fp);
-	utf8pad(buf, sizeof(buf), ci->summary ? ci->summary : "", 50, ' ');
+	utf8pad(buf, sizeof(buf), ci->summary ? ci->summary : "", 40, ' ');
 	gphlink(fp, buf, strlen(buf));
 	fputs("  ", fp);
-	utf8pad(buf, sizeof(buf), ci->author ? ci->author->name : "", 25, ' ');
+	utf8pad(buf, sizeof(buf), ci->author ? ci->author->name : "", 19, '\0');
 	gphlink(fp, buf, strlen(buf));
-	fprintf(fp, "|%s/commit/%s.gph", relpath, ci->oid);
-	fputs("|server|port]\n", fp);
+	fprintf(fp, "|%s/commit/%s.gph|server|port]\n", relpath, ci->oid);
 }
 
 int
@@ -662,32 +830,46 @@ writelog(FILE *fp, const git_oid *oid)
 	struct commitinfo *ci;
 	git_revwalk *w = NULL;
 	git_oid id;
-	char path[PATH_MAX];
+	char path[PATH_MAX], oidstr[GIT_OID_HEXSZ + 1];
 	FILE *fpfile;
 	int r;
 
 	git_revwalk_new(&w, repo);
 	git_revwalk_push(w, oid);
-	git_revwalk_sorting(w, GIT_SORT_TIME);
-	git_revwalk_simplify_first_parent(w);
 
 	while (!git_revwalk_next(&id, w)) {
 		if (cachefile && !memcmp(&id, &lastoid, sizeof(id)))
 			break;
+
+		git_oid_tostr(oidstr, sizeof(oidstr), &id);
+		r = snprintf(path, sizeof(path), "commit/%s.gph", oidstr);
+		if (r < 0 || (size_t)r >= sizeof(path))
+			errx(1, "path truncated: 'commit/%s.gph'", oidstr);
+		r = access(path, F_OK);
+
+		/* optimization: if there are no log lines to write and
+		   the commit file already exists: skip the diffstat */
+		if (!nlogcommits && !r)
+			continue;
+
 		if (!(ci = commitinfo_getbyoid(&id)))
 			break;
 
-		writelogline(fp, ci);
+		if (nlogcommits < 0) {
+			writelogline(fp, ci);
+		} else if (nlogcommits > 0) {
+			writelogline(fp, ci);
+			nlogcommits--;
+			if (!nlogcommits && ci->parentoid[0])
+				fprintf(fp, "%16.16s  More commits remaining [...]\n", "");
+		}
+
 		if (cachefile)
 			writelogline(wcachefp, ci);
 
-		r = snprintf(path, sizeof(path), "commit/%s.gph", ci->oid);
-		if (r == -1 || (size_t)r >= sizeof(path))
-			errx(1, "path truncated: 'commit/%s.gph'", ci->oid);
-
 		/* check if file exists if so skip it */
-		if (access(path, F_OK)) {
-			/* lookup stats: only required here */
+		if (r) {
+			/* lookup stats: only required here for gopher */
 			if (commitinfo_getstats(ci) == -1)
 				goto err;
 
@@ -706,7 +888,7 @@ err:
 }
 
 void
-printcommitatom(FILE *fp, struct commitinfo *ci)
+printcommitatom(FILE *fp, struct commitinfo *ci, const char *tag)
 {
 	fputs("<entry>\n", fp);
 
@@ -723,14 +905,19 @@ printcommitatom(FILE *fp, struct commitinfo *ci)
 	}
 	if (ci->summary) {
 		fputs("<title type=\"text\">", fp);
+		if (tag && tag[0]) {
+			fputs("[", fp);
+			xmlencode(fp, tag, strlen(tag));
+			fputs("] ", fp);
+		}
 		xmlencode(fp, ci->summary, strlen(ci->summary));
 		fputs("</title>\n", fp);
 	}
-	fprintf(fp, "<link rel=\"alternate\" type=\"text/html\" href=\"commit/%s.gph\" />",
-	        ci->oid);
+	fprintf(fp, "<link rel=\"alternate\" href=\"%scommit/%s.gph\" />\n",
+	        baseurl, ci->oid);
 
 	if (ci->author) {
-		fputs("<author><name>", fp);
+		fputs("<author>\n<name>", fp);
 		xmlencode(fp, ci->author->name, strlen(ci->author->name));
 		fputs("</name>\n<email>", fp);
 		xmlencode(fp, ci->author->email, strlen(ci->author->email));
@@ -748,18 +935,20 @@ printcommitatom(FILE *fp, struct commitinfo *ci)
 		xmlencode(fp, ci->author->email, strlen(ci->author->email));
 		fputs("&gt;\nDate:   ", fp);
 		printtime(fp, &(ci->author->when));
-		fputc('\n', fp);
+		putc('\n', fp);
 	}
 	if (ci->msg) {
-		fputc('\n', fp);
+		putc('\n', fp);
 		xmlencode(fp, ci->msg, strlen(ci->msg));
 	}
 	fputs("\n</content>\n</entry>\n", fp);
 }
 
 int
-writeatom(FILE *fp)
+writeatom(FILE *fp, int all)
 {
+	struct referenceinfo *ris = NULL;
+	size_t refcount = 0;
 	struct commitinfo *ci;
 	git_revwalk *w = NULL;
 	git_oid id;
@@ -772,29 +961,40 @@ writeatom(FILE *fp)
 	xmlencode(fp, description, strlen(description));
 	fputs("</subtitle>\n", fp);
 
-	git_revwalk_new(&w, repo);
-	git_revwalk_push_head(w);
-	git_revwalk_sorting(w, GIT_SORT_TIME);
-	git_revwalk_simplify_first_parent(w);
+	/* all commits or only tags? */
+	if (all) {
+		git_revwalk_new(&w, repo);
+		git_revwalk_push_head(w);
+		for (i = 0; i < m && !git_revwalk_next(&id, w); i++) {
+			if (!(ci = commitinfo_getbyoid(&id)))
+				break;
+			printcommitatom(fp, ci, "");
+			commitinfo_free(ci);
+		}
+		git_revwalk_free(w);
+	} else if (getrefs(&ris, &refcount) != -1) {
+		/* references: tags */
+		for (i = 0; i < refcount; i++) {
+			if (git_reference_is_tag(ris[i].ref))
+				printcommitatom(fp, ris[i].ci,
+				                git_reference_shorthand(ris[i].ref));
 
-	for (i = 0; i < m && !git_revwalk_next(&id, w); i++) {
-		if (!(ci = commitinfo_getbyoid(&id)))
-			break;
-		printcommitatom(fp, ci);
-		commitinfo_free(ci);
+			commitinfo_free(ris[i].ci);
+			git_reference_free(ris[i].ref);
+		}
+		free(ris);
 	}
-	git_revwalk_free(w);
 
 	fputs("</feed>\n", fp);
 
 	return 0;
 }
 
-int
-writeblob(git_object *obj, const char *fpath, const char *filename, git_off_t filesize)
+size_t
+writeblob(git_object *obj, const char *fpath, const char *filename, size_t filesize)
 {
 	char tmp[PATH_MAX] = "", *d;
-	int lc = 0;
+	size_t lc = 0;
 	FILE *fp;
 
 	if (strlcpy(tmp, fpath, sizeof(tmp)) >= sizeof(tmp))
@@ -806,9 +1006,9 @@ writeblob(git_object *obj, const char *fpath, const char *filename, git_off_t fi
 
 	fp = efopen(fpath, "w");
 	writeheader(fp, filename);
-	fputc('t', fp);
+	putc('t', fp);
 	gphtext(fp, filename, strlen(filename));
-	fprintf(fp, " (%juB)\n", (uintmax_t)filesize);
+	fprintf(fp, " (%zuB)\n", filesize);
 	fputs("---\n", fp);
 
 	if (git_blob_is_binary((git_blob *)obj)) {
@@ -870,13 +1070,11 @@ int
 writefilestree(FILE *fp, git_tree *tree, const char *path)
 {
 	const git_tree_entry *entry = NULL;
-	git_submodule *module = NULL;
 	git_object *obj = NULL;
-	git_off_t filesize;
 	const char *entryname;
-	char buf[256], filepath[PATH_MAX], entrypath[PATH_MAX];
-	size_t count, i;
-	int lc, r, ret;
+	char buf[256], filepath[PATH_MAX], entrypath[PATH_MAX], oid[8];
+	size_t count, i, lc, filesize;
+	int r, ret;
 
 	count = git_tree_entrycount(tree);
 	for (i = 0; i < count; i++) {
@@ -887,7 +1085,7 @@ writefilestree(FILE *fp, git_tree *tree, const char *path)
 
 		r = snprintf(filepath, sizeof(filepath), "file/%s.gph",
 		         entrypath);
-		if (r == -1 || (size_t)r >= sizeof(filepath))
+		if (r < 0 || (size_t)r >= sizeof(filepath))
 			errx(1, "path truncated: 'file/%s.gph'", entrypath);
 
 		if (!git_tree_entry_to_object(&obj, repo, entry)) {
@@ -917,19 +1115,22 @@ writefilestree(FILE *fp, git_tree *tree, const char *path)
 			gphlink(fp, buf, strlen(buf));
 			fputs("  ", fp);
 			if (lc > 0)
-				fprintf(fp, "%7dL", lc);
+				fprintf(fp, "%7zuL", lc);
 			else
-				fprintf(fp, "%7juB", (uintmax_t)filesize);
-			fprintf(fp, "|%s/%s", relpath, filepath);
+				fprintf(fp, "%7zuB", filesize);
+			fprintf(fp, "|%s/", relpath);
+			gphlink(fp, filepath, strlen(filepath));
 			fputs("|server|port]\n", fp);
 			git_object_free(obj);
-		} else if (!git_submodule_lookup(&module, repo, entryname)) {
+		} else if (git_tree_entry_type(entry) == GIT_OBJ_COMMIT) {
+			/* commit object in tree is a submodule */
 			fputs("[1|m---------  ", fp);
-			utf8pad(buf, sizeof(buf), entrypath, 50, ' ');
-			gphlink(fp, buf, strlen(buf));
+			gphlink(fp, entrypath, strlen(entrypath));
+			fputs(" @ ", fp);
+			git_oid_tostr(oid, sizeof(oid), git_tree_entry_id(entry));
+			gphlink(fp, oid, strlen(oid));
 			fprintf(fp, "|%s/file/.gitmodules.gph|server|port]\n", relpath);
 			/* NOTE: linecount omitted */
-			git_submodule_free(module);
 		}
 	}
 
@@ -957,116 +1158,63 @@ writefiles(FILE *fp, const git_oid *id)
 	return ret;
 }
 
-int
-refs_cmp(const void *v1, const void *v2)
-{
-	git_reference *r1 = (*(git_reference **)v1);
-	git_reference *r2 = (*(git_reference **)v2);
-	int r;
-
-	if ((r = git_reference_is_branch(r1) - git_reference_is_branch(r2)))
-		return r;
-
-	return strcmp(git_reference_shorthand(r1),
-	              git_reference_shorthand(r2));
-}
-
 int
 writerefs(FILE *fp)
 {
+	struct referenceinfo *ris = NULL;
 	struct commitinfo *ci;
-	const git_oid *id = NULL;
-	git_object *obj = NULL;
-	git_reference *dref = NULL, *r, *ref = NULL;
-	git_reference_iterator *it = NULL;
-	git_reference **refs = NULL;
 	size_t count, i, j, refcount;
 	const char *titles[] = { "Branches", "Tags" };
-	const char *name;
+	const char *s;
 	char buf[256];
 
-	if (git_reference_iterator_new(&it, repo))
+	if (getrefs(&ris, &refcount) == -1)
 		return -1;
 
-	for (refcount = 0; !git_reference_next(&ref, it); refcount++) {
-		if (!(refs = reallocarray(refs, refcount + 1, sizeof(git_reference *))))
-			err(1, "realloc");
-		refs[refcount] = ref;
-	}
-	git_reference_iterator_free(it);
-
-	/* sort by type then shorthand name */
-	qsort(refs, refcount, sizeof(git_reference *), refs_cmp);
-
-	for (j = 0; j < 2; j++) {
-		for (i = 0, count = 0; i < refcount; i++) {
-			if (!(git_reference_is_branch(refs[i]) && j == 0) &&
-			    !(git_reference_is_tag(refs[i]) && j == 1))
-				continue;
-
-			switch (git_reference_type(refs[i])) {
-			case GIT_REF_SYMBOLIC:
-				if (git_reference_resolve(&dref, refs[i]))
-					goto err;
-				r = dref;
-				break;
-			case GIT_REF_OID:
-				r = refs[i];
-				break;
-			default:
-				continue;
-			}
-			if (!git_reference_target(r) ||
-			    git_reference_peel(&obj, r, GIT_OBJ_ANY))
-				goto err;
-			if (!(id = git_object_id(obj)))
-				goto err;
-			if (!(ci = commitinfo_getbyoid(id)))
-				break;
-
-			/* print header if it has an entry (first). */
-			if (++count == 1) {
-				fprintf(fp, "%s\n", titles[j]);
-				fprintf(fp, "  %-20.20s", "Name");
-				fprintf(fp, "  %-16.16s", "Last commit date");
-				fprintf(fp, "  %s\n", "Author");
-			}
-
-			name = git_reference_shorthand(r);
-
-			fputs("  ", fp);
-			utf8pad(buf, sizeof(buf), name, 20, ' ');
-			gphlink(fp, buf, strlen(buf));
-			fputs("  ", fp);
-			if (ci->author)
-				printtimeshort(fp, &(ci->author->when));
-			else
-				fputs("                ", fp);
-			fputs("  ", fp);
-			if (ci->author) {
-				utf8pad(buf, sizeof(buf), ci->author->name, 25, ' ');
-				gphlink(fp, buf, strlen(buf));
-			}
-			fputs("\n", fp);
-
-			commitinfo_free(ci);
-			git_object_free(obj);
-			obj = NULL;
-			git_reference_free(dref);
-			dref = NULL;
+	for (i = 0, j = 0, count = 0; i < refcount; i++) {
+		if (j == 0 && git_reference_is_tag(ris[i].ref)) {
+			/* table footer */
+			if (count)
+				fputs("\n", fp);
+			count = 0;
+			j = 1;
 		}
-		/* table footer */
-		if (count)
-			fputs("\n", fp);
+
+		/* print header if it has an entry (first). */
+		if (++count == 1) {
+			fprintf(fp, "%s\n", titles[j]);
+			fprintf(fp, "  %-32.32s", "Name");
+			fprintf(fp, "  %-16.16s", "Last commit date");
+			fprintf(fp, "  %s\n", "Author");
+		}
+
+		ci = ris[i].ci;
+		s = git_reference_shorthand(ris[i].ref);
+
+		fputs("  ", fp);
+		utf8pad(buf, sizeof(buf), s, 32, ' ');
+		gphlink(fp, buf, strlen(buf));
+		fputs("  ", fp);
+		if (ci->author)
+			printtimeshort(fp, &(ci->author->when));
+		else
+			fputs("                ", fp);
+		fputs("  ", fp);
+		if (ci->author) {
+			utf8pad(buf, sizeof(buf), ci->author->name, 25, '\0');
+			gphlink(fp, buf, strlen(buf));
+		}
+		fputs("\n", fp);
 	}
+	/* table footer */
+	if (count)
+		fputs("\n", fp);
 
-err:
-	git_object_free(obj);
-	git_reference_free(dref);
-
-	for (i = 0; i < refcount; i++)
-		git_reference_free(refs[i]);
-	free(refs);
+	for (i = 0; i < refcount; i++) {
+		commitinfo_free(ris[i].ci);
+		git_reference_free(ris[i].ref);
+	}
+	free(ris);
 
 	return 0;
 }
@@ -1074,7 +1222,8 @@ err:
 void
 usage(char *argv0)
 {
-	fprintf(stderr, "%s [-b baseprefix] [-c cachefile] repodir\n", argv0);
+	fprintf(stderr, "%s [-b baseprefix] [-c cachefile | -l commits] "
+	        "[-u baseurl] repodir\n", argv0);
 	exit(1);
 }
 
@@ -1083,7 +1232,6 @@ main(int argc, char *argv[])
 {
 	git_object *obj = NULL;
 	const git_oid *head = NULL;
-	const git_error *e = NULL;
 	mode_t mask;
 	FILE *fp, *fpread;
 	char path[PATH_MAX], repodirabs[PATH_MAX + 1], *p;
@@ -1091,9 +1239,6 @@ main(int argc, char *argv[])
 	size_t n;
 	int i, fd;
 
-	if (pledge("stdio rpath wpath cpath fattr", NULL) == -1)
-		err(1, "pledge");
-
 	setlocale(LC_CTYPE, "");
 
 	for (i = 1; i < argc; i++) {
@@ -1106,13 +1251,23 @@ main(int argc, char *argv[])
 				usage(argv[0]);
 			relpath = argv[++i];
 		} else if (argv[i][1] == 'c') {
-			if (i + 1 >= argc)
+			if (nlogcommits > 0 || i + 1 >= argc)
 				usage(argv[0]);
 			cachefile = argv[++i];
+		} else if (argv[i][1] == 'l') {
+			if (cachefile || i + 1 >= argc)
+				usage(argv[0]);
+			errno = 0;
+			nlogcommits = strtoll(argv[++i], &p, 10);
+			if (argv[i][0] == '\0' || *p != '\0' ||
+			    nlogcommits <= 0 || errno)
+				usage(argv[0]);
+		} else if (argv[i][1] == 'u') {
+			if (i + 1 >= argc)
+				usage(argv[0]);
+			baseurl = argv[++i];
 		}
 	}
-	if (!cachefile && pledge("stdio rpath wpath cpath", NULL) == -1)
-		err(1, "pledge");
 	if (!repodir)
 		usage(argv[0]);
 
@@ -1121,10 +1276,26 @@ main(int argc, char *argv[])
 
 	git_libgit2_init();
 
+#ifdef __OpenBSD__
+	if (unveil(repodir, "r") == -1)
+		err(1, "unveil: %s", repodir);
+	if (unveil(".", "rwc") == -1)
+		err(1, "unveil: .");
+	if (cachefile && unveil(cachefile, "rwc") == -1)
+		err(1, "unveil: %s", cachefile);
+
+	if (cachefile) {
+		if (pledge("stdio rpath wpath cpath fattr", NULL) == -1)
+			err(1, "pledge");
+	} else {
+		if (pledge("stdio rpath wpath cpath", NULL) == -1)
+			err(1, "pledge");
+	}
+#endif
+
 	if (git_repository_open_ext(&repo, repodir,
 		GIT_REPOSITORY_OPEN_NO_SEARCH, NULL) < 0) {
-		e = giterr_last();
-		fprintf(stderr, "%s: %s\n", argv[0], e->message);
+		fprintf(stderr, "%s: cannot open repository\n", argv[0]);
 		return 1;
 	}
 
@@ -1172,17 +1343,24 @@ main(int argc, char *argv[])
 	}
 
 	/* check LICENSE */
-	haslicense = (!git_revparse_single(&obj, repo, "HEAD:LICENSE") &&
-		git_object_type(obj) == GIT_OBJ_BLOB);
-	git_object_free(obj);
+	for (i = 0; i < LEN(licensefiles) && !license; i++) {
+		if (!git_revparse_single(&obj, repo, licensefiles[i]) &&
+		    git_object_type(obj) == GIT_OBJ_BLOB)
+			license = licensefiles[i] + strlen("HEAD:");
+		git_object_free(obj);
+	}
 
 	/* check README */
-	hasreadme = (!git_revparse_single(&obj, repo, "HEAD:README") &&
-		git_object_type(obj) == GIT_OBJ_BLOB);
-	git_object_free(obj);
+	for (i = 0; i < LEN(readmefiles) && !readme; i++) {
+		if (!git_revparse_single(&obj, repo, readmefiles[i]) &&
+		    git_object_type(obj) == GIT_OBJ_BLOB)
+			readme = readmefiles[i] + strlen("HEAD:");
+		git_object_free(obj);
+	}
 
-	hassubmodules = (!git_revparse_single(&obj, repo, "HEAD:.gitmodules") &&
-		git_object_type(obj) == GIT_OBJ_BLOB);
+	if (!git_revparse_single(&obj, repo, "HEAD:.gitmodules") &&
+	    git_object_type(obj) == GIT_OBJ_BLOB)
+		submodules = ".gitmodules";
 	git_object_free(obj);
 
 	/* log for HEAD */
@@ -1191,7 +1369,7 @@ main(int argc, char *argv[])
 	writeheader(fp, "Log");
 
 	fprintf(fp, "%-16.16s  ", "Date");
-	fprintf(fp, "%-50.50s  ", "Commit message");
+	fprintf(fp, "%-40.40s  ", "Commit message");
 	fprintf(fp, "%s\n", "Author");
 
 	if (cachefile && head) {
@@ -1231,7 +1409,9 @@ main(int argc, char *argv[])
 		if (head)
 			writelog(fp, head);
 	}
-	fprintf(fp, "\n[0|Atom feed|%s/atom.xml|server|port]\n", relpath);
+	fputs("\n", fp);
+	fprintf(fp, "[0|Atom feed|%s/atom.xml|server|port]\n", relpath);
+	fprintf(fp, "[0|Atom feed (tags)|%s/tags.xml|server|port]\n", relpath);
 	writefooter(fp);
 	fclose(fp);
 
@@ -1252,7 +1432,12 @@ main(int argc, char *argv[])
 
 	/* Atom feed */
 	fp = efopen("atom.xml", "w");
-	writeatom(fp);
+	writeatom(fp, 1);
+	fclose(fp);
+
+	/* Atom feed for tags / releases */
+	fp = efopen("tags.xml", "w");
+	writeatom(fp, 0);
 	fclose(fp);
 
 	/* rename new cache file on success */
